package net.udby.datetime.derby;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Scanner;
import java.util.TreeMap;

import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;

import net.udby.datetime.jpa.Item;

/**
 * Helper class for unit-tests using in-memory derby instance
 */
public class DbTestHelper {
	public static final String PERSISTENCE_CONTEXT_NAME = "derby-inmemory-test";

	public static final String TESTDB_DRIVER = "org.apache.derby.jdbc.EmbeddedDriver";
	public static final String TESTDB_URL = "jdbc:derby:memory:test;create=true";
	public static final String TESTDB_USER = "user";
	public static final String TESTDB_PW = "password";

	public static final String TEARDOWN_SQL = "/sql/teardown.sql";
	public static final String SETUP_SQL = "/sql/setup.sql";
	
	/** List of tables containing "live" information that should be scrapped before each method */
	private static final String[] LIVE_TABLES = { Item.TABLE_NAME };

	private static DbTestHelper instance;
	
	public final Connection connection;
	public final EntityManagerFactory entityManagerFactory;

	private DbTestHelper(final Connection connection, final EntityManagerFactory entityManagerFactory) {
		this.connection = connection;
		this.entityManagerFactory = entityManagerFactory;
	}
	public static DbTestHelper create(String driver, String url, String user, String password, String persistenceUnit) {
		try {
			Class.forName(driver);
			// note: this connection is kept alive for as long as the jvm is alive (it is not closed); this is ok for simple test-purposes...
			Connection connection = DriverManager.getConnection(url, user, password);
			
			Map<String, String> props = new HashMap<String, String>();
			props.put("javax.persistence.jdbc.driver", driver);
			props.put("javax.persistence.jdbc.url", url);
			props.put("javax.persistence.jdbc.user", user);
			props.put("javax.persistence.jdbc.password", password);

			return new DbTestHelper(connection, Persistence.createEntityManagerFactory(persistenceUnit, props));
		} catch (Exception e) {
			throw new RuntimeException("Error setting up test-database-environment", e);
		}		
	}
	
	public static void initialize() {
		if (instance == null) {
			DbTestHelper helper = create(TESTDB_DRIVER, TESTDB_URL, TESTDB_USER, TESTDB_PW, PERSISTENCE_CONTEXT_NAME);
			try {
				runScript(helper.connection, TEARDOWN_SQL, true, true);

				helper.connection.clearWarnings();

				runScript(helper.connection, SETUP_SQL, false, true);

				helper.connection.clearWarnings();
				
				instance = helper;
			} catch (Exception e) {
				throw new RuntimeException("Error setting up test-database-environment", e);
			}
		}
	}

	public static void truncate(String tableName) {
		try {
			PreparedStatement ps = instance.connection.prepareStatement("DELETE FROM " + tableName);
			try {
				ps.executeUpdate();
				instance.connection.commit();
			} finally {
				ps.close();
			}
		} catch (SQLException e) {
			throw new RuntimeException("Truncate failed", e);
		}
	}
	
	public static void truncateAll() {
		for (String tableName : LIVE_TABLES) {
			truncate(tableName);
		}
	}
	
	public static Map<String, Object> selectRow(String tableName, String whereClause, boolean checkUnique, Object... args) {
		return selectRow(instance.connection, tableName, whereClause, checkUnique, args);
	}
	
	public static Map<String, Object> selectRow(Connection connection, String tableName, String whereClause, boolean checkUnique, Object... args) {
		String query = "SELECT * FROM " + tableName + " WHERE " + whereClause;
		try {
			PreparedStatement ps = connection.prepareStatement(query);
			try {
				for (int i = 0; i < args.length; i++) {
					ps.setObject(i+1, args[i]);
				}
				ResultSet rs = ps.executeQuery();
				try {
					if (rs.next()) {
						Map<String, Object> result = new TreeMap<String, Object>();
						ResultSetMetaData rsm = rs.getMetaData();
						for (int i = 0; i < rsm.getColumnCount(); i++) {
							String column = rsm.getColumnLabel(i+1);
							Object value = rs.getObject(i+1);
							result.put(column, value);
						}
						if (checkUnique && rs.next()) {
							throw new RuntimeException("Result is not unique - there is more than one");
						}
						return result;
					} else {
						if (checkUnique) {
							throw new RuntimeException("Result is not unique - there are none");
						}
						return Collections.emptyMap();
					}
				} finally {
					rs.close();
				}
			} finally {
				ps.close();
			}
		} catch (SQLException e) {
			throw new RuntimeException("Query["+query+"] failed", e);
		}
	}
	
	public static EntityManagerFactory getEntityManagerFactory() {
		return instance.entityManagerFactory;
	}
	
	public static void runScript(String sqlResource, boolean silent, boolean trim) throws Exception {
		runScript(instance.connection, sqlResource, silent, trim);
	}
	
	public static void runScript(Connection connection, String sqlResource, boolean silent, boolean trim) throws Exception {
		InputStream is = DbTestHelper.class.getResourceAsStream(sqlResource);
		if (is == null) {
			throw new RuntimeException("Cannot run script. Cannot load "+sqlResource+" from classpath");
		}
		try {
			Scanner sc = new Scanner(new InputStreamReader(is, "UTF-8"));
			StringBuilder sql = new StringBuilder();
			for (; sc.hasNextLine();) {
				String line = sc.nextLine();
				// remove comments - note this simple approach only works if valid strings does not contain '--'
				int cpos = line.indexOf("--");
				if (cpos >= 0) {
					line = line.substring(0, cpos);
				}
				if (trim) {
					// trim...
					line = line.trim();
					if (!line.isEmpty()) {
						sql.append(line);
					}
				} else {
					sql.append(line);
					if (sc.hasNextLine()) {
						sql.append("\n");
					}
				}
			}
			sc.close();
			
			Statement stmt;
			try {
				stmt = connection.createStatement();
			} catch (Exception e) {
				throw new RuntimeException("cannot create statement", e);
			}
			String s = sql.toString();
			try {
				execute(stmt, s, silent);
			} finally {
				try {
					connection.commit();
				} catch (Exception ignored) {
				}
				try {
					stmt.close();
				} catch (Exception ignored) {
				}
			}
		} finally {
			try {
				is.close();
			} catch (IOException ignored) {
			}
		}
	}

	public static void execute(Statement stmt, String sqls, boolean silent) throws SQLException {
		String[] parts = sqls.split(";");
		for (String sql : parts) {
			if (!sql.trim().isEmpty()) {
				try {
					stmt.execute(sql);
				} catch (SQLException e) {
					//System.err.println("Possible problem (" + e.toString() + ") with sql='"+sql+"'");
					if (!silent) throw e;
				}
			}
		}
	}

}
